
#include <iostream.h>
#include <stdiostream.h>
#include <fstream.h>

#include "ECArgs.h"
#include "utils.h"
#include "InputTree.h"
#include "fixSent.h"
#include "Usi.h"
#include "headFinder.h"
#include <bstring.h>
#include "Term.h"
#include <rw/tpslist.h>

/* read through wsj training data.  for each tagged-sentence/rule pair
   count how often we see a nonterm X preceeded by tag Y, and how many
   tags Y we see, then p(X|Y) = |X, Y|/|Y| */

#define Termstar const Term*
FILE * 		    resultsFP = NULL;

int cNtt[155][50];
int ct[205];
int ctNt[155][50];
int c = 0;

int* agltconta_[155][205];  //155 possible ls, 205 possible ts, 50 possible a's
int  ltconta[155][205];
//int* bgltcontb_[155][205];  //155 possible ls, 205 possible ts, 50 possible a's
//int  btcont[50][205];
int  lgbcont[50][155];
int* aglh_[155][205];
int  lh[155][205];


InputTree* sentence[256];
int endPos;
void wordsFromTree(InputTree* tree);

void
makeSent(InputTree* tree)
{
  endPos = 0;
  for(int i = 0 ; i < 128 ; i++) sentence[i] = NULL;
  wordsFromTree(tree);
  assert(endPos == tree->finish());
}

void
wordsFromTree(InputTree* tree)
{
  if(tree->word() != "")
    {
      sentence[endPos++] = tree;
      return;
    }
  InputTreesIter subTreeIter = tree->subTrees().begin();
  for( ; subTreeIter != tree->subTrees().end() ; subTreeIter++ )
    {
      InputTree* subTree = *subTreeIter;
      wordsFromTree(subTree);
    }
}

/* counts up how often we see nt above t followed by b in the same constit
   where b is a preterminal */

void
genIncrAgtlconta(int* gAgltconta[155][205], int l, int t, int a)
{
  int* ctNt = gAgltconta[l][t];
  if(!ctNt)
    {
      ctNt = new int[50];
      for(int i = 0 ; i < 50 ; i++) ctNt[i] = 0;
      gAgltconta[l][t] = ctNt;
    }
  ctNt[a]++;
}

void
incrAgltconta(int l, int t, int a)
{
  genIncrAgtlconta(agltconta_, l, t, a);
}

void
incrAglh(int l, int t, int b)
{
  genIncrAgtlconta(aglh_, l, t, b);
}


int
genAgltconta(int* gagltconta[155][205], int l, int t, int a)
{
  int* ctNt = gagltconta[l][t];
  if(!ctNt) return 0;
  return ctNt[a];
}

int
agltconta(int l, int t, int a)
{
  return genAgltconta(agltconta_, l, t, a);
}
/*
int
bgltcontb(int l, int t, int b)
{
  return genAgltconta(bgltcontb_, l, t, b);
}
*/
int
aglh(int l, int t, int b)
{
  return genAgltconta(aglh_, l, t, b);
}


string
findAcTerm(int pos)
{
  if(pos < 0) return "STOP";
  if(pos >= endPos) return "STOP";
  else return sentence[pos]->term();
}
    
void
countNtt(InputTree* tree, int wordPos)
{
  //cerr << tree->start() << " " << tree->finish() << " " << *tree << endl;
  bool okSit = true;

  if( tree->word() != ""  )
    {
      const Term* trm = Term::get(tree->term());
      int trmInt = trm->toInt();
      ct[trmInt]++;
      c++;
      return;
    }
  string fixedTerm(tree->term());
  if(fixedTerm == "")                     //??? a hack until I decide.
    fixedTerm = "S1";
  const Term* lhs = Term::get(fixedTerm);
  /* If we cannot recognize the term, don't abort, just warn and do not
     create a rule here or one level up. */
  int lhsInt = -1;
  if(lhs)
    {
      int fullLhsInt =  lhs->toInt();
      lhsInt = fullLhsInt - Term::lastTagInt() - 1;
      assert(lhsInt < 155);
      if(lhs->terminal_p()) return;
      ct[fullLhsInt]++;  // used for p(t|Nt);
      const Term* prevTerm;
      prevTerm = Term::get(findAcTerm(wordPos-1));
      //cerr << *lhs << " " << *prevTerm << " " << wordPos << endl;
      if(prevTerm)
	{
	  int pTermInt = prevTerm->toInt();
	  cNtt[lhsInt][pTermInt]++;
	}
    }
  InputTreesIter subTreeIter = tree->subTrees().begin();
  InputTree  *subTree;
  int subTlen = tree->subTrees().size();
  int subTpos = 0;
  bool badSit = false;
  RWTPtrSlist<ConstTerm> rhs;
  for( ; subTreeIter != tree->subTrees().end() ; subTreeIter++)
    {
      subTpos++;
      subTree = *(subTreeIter);
      ConstTerm* stt = Term::get(subTree->term());
      if(!stt)
	{
	  badSit = true;
	  continue;
	}
      rhs.insert(stt);
      countNtt(subTree, wordPos);
      wordPos = subTree->finish();
    }
  if(!lhs) return;
  if(badSit) return;
  if(lhs)
    {
      const Term* nxtTerm;
      nxtTerm = Term::get(findAcTerm(wordPos));
      if(nxtTerm)
	{
	  int nTermInt = nxtTerm->toInt();
	  ctNt[lhsInt][nTermInt]++;
	}
    }

  int rhsLen = tree->subTrees().size();

  int headpos = headPosFromTree(tree);
  InputTree* headTree = ithInputTree(headpos,tree->subTrees());
  int after;
  int headInt = Term::get(headTree->term())->toInt();
  ConstTerm* ahTerm = Term::get(findAcTerm(headTree->finish()));
  if(ahTerm)
    {
      int ahInt = ahTerm->toInt();
      incrAglh(lhsInt, headInt, ahInt);
      lh[lhsInt][headInt]++;
    }
  int relpos;
  for(relpos = 0 ; relpos + headpos > 0 ; relpos--)
    {
      int abpos = relpos + headpos;
      InputTree* relTree = ithInputTree(abpos,tree->subTrees());
      int bPos = relTree->start() -1;
      ConstTerm* bTerm = Term::get(findAcTerm(bPos));
      if(!bTerm) continue;
      int bInt = bTerm->toInt();
      if(bTerm->name() == "STOP") continue;  //this can happen because
	//fixSent can make the start of the sentence later than is
	  //indicated by relTree;
      assert( bInt >= 0 && bInt < 50 );
      lgbcont[bInt][lhsInt]++;
    }
  for(relpos = 1; relpos + headpos < rhsLen ; relpos++)
    {
      int abpos = relpos + headpos;
      InputTree* relTree = ithInputTree(abpos,tree->subTrees());
      int aPos = relTree->start();
      ConstTerm* aTerm = Term::get(findAcTerm(aPos));
      if(!aTerm) continue;
      int aInt = aTerm->toInt();
      ConstTerm* tTerm = rhs[abpos-1];
      if(!tTerm) continue;
      int tInt = tTerm->toInt();
      incrAgltconta(lhsInt, tInt, aInt); //???;
      //cerr << "incrA " << lhsInt+Term::lastTagInt() + 1 << " "
	//<< tInt << " " << aInt << endl;
      ltconta[lhsInt][tInt]++;
    }
  return;
}

int
main(int argc, char *argv[])
{
  ECArgs args( argc, argv);
  string path(args.arg(0));
  cout << "At start of prbnttn" << endl;

  string trainingString(path);
  trainingString += "train.txt";
  ifstream    trainingStream(trainingString);
  assert(trainingStream);

  Term::init( path );  
  readHeadInfo(path);

  const Term* gTerm = Term::get("STOP");//
  int gTint = gTerm->toInt();
  int i, j, k;
  for(i = 0 ; i < 205 ; i++)
    {
      if(i >= 155) continue;
      for(j = 0 ; j < 205 ; j++)
	{
	  agltconta_[i][j] = NULL;
	  ltconta[i][j] = 0;
	  //bgltcontb_[i][j] = NULL;
	  if(j >= 50) continue;
	  cNtt[i][j] = 0;
	  ctNt[i][j] = 0;
	}
    }

  for(i = 0 ; i < 50 ; i++)
    for(j = 0 ; j < 155 ; j++) 
      lgbcont[i][j] = 0;
  for(i = 0 ; i < 155 ; i++)
    for(j = 0 ; j < 205 ; j++) 
      lh[i][j] = 0;
  
  for(i = 0 ; i < 205 ; i++) ct[i] = 0;

  int topInt = Term::get("S1")->toInt() - Term::lastTagInt() -1;
  cerr<<"before training\n";
  i = 0;
  while(trainingStream)
    {
      //if(i > 1000) break;
      if(i++%5000 == 0) cerr << i << endl;
      InputTree  parse;
      trainingStream >> parse;
      if(!trainingStream) break;

      EcSPairs wtList;
      parse.make(wtList); 
      int endRemoved = 0;
      int numRemoved = 0;
      fixSent(wtList, numRemoved, endRemoved);
      InputTree* par;
      if(numRemoved + endRemoved)
	{
	  //cerr << "Fix info: " << numRemoved << " " << endRemoved << endl;
	  //cerr << parse << endl;
	  par=unfixParse(numRemoved, endRemoved, &parse, parse.finish());
	  //if(par) cerr << *par << endl;
	  if(!par)
	    {
	      par = &parse;
	      numRemoved = 0;
	      endRemoved = 0;
	    }
	}
      else par = &parse;

      ct[gTint]++; // STOP is eosent marker;
      /* sent is a list of alternating strings of words and parts of speech. */
      makeSent(par);
      countNtt(par, 0);
      if(numRemoved + endRemoved) delete par;
    }
  //cerr<<"after training data\n";

  string resultsString(path);
  resultsString += "nttprobs.txt";
  ofstream  resultsStream(resultsString);
  assert(resultsStream);

  // first we print out p(N|t)s;
  for(i = 0 ; i < 154 ; i++)
    {
      //cerr<<i<<endl;
      int actI = Term::lastTagInt() + 1 + i;
      Termstar trm = Term::fromInt(actI);
      //cerr<<trm->name()<<endl;
      if(trm->name() == "G1") break;
      // print N ;
      resultsStream << trm->name() << "\n";
      for(j = 0 ; j < 50 ; j++)
	{
	  //cerr<<i<<' '<<j<<endl;
	  float denom = ct[j];
	  if(denom == 0) continue;
	  int num = cNtt[i][j];
	  if(num == 0) continue;
	  Termstar strm = Term::fromInt(j);
	  // print t 
	       resultsStream << "\t" << strm->name() << "\t";
	  // print p(N|t) = |N, t| / |t| ;
	  resultsStream << (float)num/(float)denom << "\n";
	}
      resultsStream << "|\n\n";
    }
  resultsStream << "|\n\n" ;
  cerr<<"first pass ok\n";
  // go through again printing p(t|NT) 
       for(i = 0 ; i < 154 ; i++)
	 {
	   //cerr<<i<<endl;
	   int actI = Term::lastTagInt() + 1 + i;
	   Termstar trm = Term::fromInt(actI);
	   if(trm->name() == "G1") break;
	   // print N ;
	   float denom = ct[actI];
	   if(denom==0) cerr<<"denom is 0 at "<<i<<endl;
	   if(denom==0) continue;
	   assert(denom);
	   resultsStream << trm->name() << "\n";
	   for(j = 0 ; j < 50 ; j++)
	     {
	       //cerr<<' '<<j<<endl;
	       Termstar strm = Term::fromInt(j);
	       // print t (= strm) 
		    float num = ctNt[i][j];
	       if(num == 0) continue;
	       resultsStream << "\t" << strm->name() << "\t";
	       resultsStream << num/denom << "\n";
	     }
	   resultsStream << "|\n\n";
	 }


  string pTcontString(path);
  pTcontString += "pTcont.txt";
  ofstream  pTcontStream(pTcontString);
  assert(pTcontStream);

  /* print out info in format,
     lhsInt   prevInt
              nextInt  p(next|...)
	      nextInt2 p(next2| ..)
	      ...
     9999 */

  for(int l = 0 ; l < 154 ; l++)
    {
      for(int t = 0 ; t < 200 ; t++)
	{
	  int denom = lh[l][t];
	  if(denom == 0) continue;
	  pTcontStream << l+Term::lastTagInt() + 1 << "\t" << t << "\n";
	  for(int nxt = 0 ; nxt < 47 ; nxt++)
	    {
	      int num = aglh(l, t, nxt);
	      if(num == 0) continue;
	      pTcontStream << "\t" << nxt << "\t"
			   << (float)num/(float)denom << "\n";
	    }
	  pTcontStream << "9999\n";
	}
    }
  pTcontStream << "\n9999\n";
  for(l = 0 ; l < 154 ; l++)
    {
      for(int t = 0 ; t < 200 ; t++)
	{
	  int denom = ltconta[l][t];
	  if(denom == 0) continue;
	  pTcontStream << l+Term::lastTagInt() + 1 << "\t" << t << "\n";
	  for(int nxt = 0 ; nxt < 47 ; nxt++)
	    {
	      int num = agltconta(l, t, nxt);
	      if(num == 0) continue;
	      pTcontStream << "\t" << nxt << "\t"
			   << (float)num/(float)denom << "\n";
	    }
	  pTcontStream << "9999\n";
	}
    }
  pTcontStream << "\n9999\n";

  for(int b = 0 ; b < 50 ; b++)
    {
      int denom = ct[b];
      if(denom == 0) continue;
      for(int l = 0 ; l < 155 ; l++)
	{
	  int num = lgbcont[b][l];
	  if(num == 0) continue;
	  pTcontStream
	    << b
	    << "\t"
	    << l+Term::lastTagInt() + 1
	    << "\t"
	    << (float)num/(float)denom
	    << "\n";
	}
    }
  return 0;
}


