
#include <fstream.h>
#include <sys/resource.h>
#include <iostream.h>
#include <unistd.h>
#include <set.h>
#include <math.h>
#include "ECArgs.h"
#include "Feature.h"
#include "FeatureTree.h"
#include "InputTree.h"
#include "headFinder.h"
#include "treeHistSf.h"
#include "Pst.h"
#include "Smoother.h"
#include "TreeHist.h"
#include "Term.h"

int pass;
int whichInt;

#define NUMBASEPASSES 5


/* for a given history, as specified by a tree, for each feature f_i record
   how often it was used. */

FeatureTree* tRoot = NULL;
string conditionedType;
InputTree* curTree = NULL;

void goThroughSents(InputTree* data[1301],int sc);

struct TrData 
{
  int c;
  float pm;
};

TrData trData[MAXNUMFS][15];

float smoothedPs[MAXNUMFS];
bool prune = false;
typedef set<string, less<string> > StringSet;
StringSet wordSet;
bool
inWordSet(int i)
{
  const string& w = Pst::fromInt(i);
  if(wordSet.find(w) == wordSet.end()) return false;
  return true;
}

InputTree* sentence[256];
int endPos;
void wordsFromTree(InputTree* tree);

void
makeSent(InputTree* tree)
{
  endPos = 0;
  for(int i = 0 ; i < 128 ; i++) sentence[i] = NULL;
  wordsFromTree(tree);
  //cerr << "tf " << tree->finish() << endl; //???;
  assert(endPos == tree->finish());
}

void
wordsFromTree(InputTree* tree)
{
  if(tree->word() != "")
    {
      sentence[endPos++] = tree;
      return;
    }
  InputTreesIter subTreeIter = tree->subTrees().begin();
  for( ; subTreeIter != tree->subTrees().end() ; subTreeIter++ )
    {
      InputTree* subTree = *subTreeIter;
      wordsFromTree(subTree);
    }
}

int nFeatVs[20];

void
processG(int i, FeatureTree* ginfo[], TreeHist* treeh, int cVal)
{
  ginfo[i] = NULL;
  Feature* feat = Feature::fromInt(i, whichInt); 
  /* e.g., g(rtlu) starts from where g(rtl) left off (after tl)*/
  int searchStartInd = feat->startPos;
  FeatureTree* strt = ginfo[searchStartInd];
  if(!strt)
    {
      if(i == 1) cerr << "no start for i = 1\n" << *curTree << endl;
      else smoothedPs[i] = smoothedPs[i-1]; //???;
      return;
    }
  SubFeature* sf = SubFeature::fromInt(feat->subFeat, whichInt);
  int nfeatV = (*(sf->fun))(treeh);
  nFeatVs[i] = nfeatV;
  //cerr << "pg " << i << ", " << cVal << ", " << nfeatV << endl;
  FeatureTree* histPt = strt->follow(nfeatV, feat->auxCnt); 
  ginfo[i] = histPt;
  if(i == 1)
    {
      smoothedPs[0] = 1;

      FeatMap::iterator fti1 = histPt->feats.find(cVal);
      float unsmoothedVal1;
      if(fti1 == histPt->feats.end()) unsmoothedVal1 = 0;
      else unsmoothedVal1 = (*fti1).second.g();

      smoothedPs[1] = unsmoothedVal1;
      if(smoothedPs[1] == 0)
	{
	  //if(pass == -1)
	  //cerr << "Zero at level 1 " << treeh->pos
	  // << "\n" << *curTree << endl;
	  smoothedPs[1] = .0001;
	}
      //cerr <<i<< " " << nfeatV << " " << cVal << " "<< smoothedPs[1] << endl;
      for(int j = 2; j <= Feature::total[whichInt] ; j++)
	smoothedPs[j] = 0;
      return;
    }
  if(!histPt)
    {
      smoothedPs[i] = smoothedPs[i-1]; //???;
      //cerr << "nhp " << i << " " << smoothedPs[i] << endl;
      return;
    }
  
  int b;
  int sz = histPt->feats.size();
  float estm = ((float)histPt->count / (float)sz);
  //float estm = histPt->count * smoothedPs[1]; //???;
  //float estm = histPt->count * 0.1;
  b = Smoother::bucket(estm,i);
  //cerr << "bu " << estm << " " << i << " " << b << endl;

  FeatMap::iterator fti = histPt->feats.find(cVal);
  float unsmoothedVal;
  if(fti == histPt->feats.end()) unsmoothedVal = 0;
  else unsmoothedVal = (*fti).second.g();
  //if(unsmoothedVal == 0)
    //cerr << "Zero at level 2 " << treeh->pos << "\n" << *curTree << endl;
  float lam = Feature::getLambda(whichInt, i, b);
  //cerr << i << " " << nfeatV << " " << histPt->featureInt << " "
  //   << estm << " " << b << " " << unsmoothedVal << endl;
  float uspathprob = lam*unsmoothedVal;
  //float osmoothedVal = smoothedPs[searchStartInd];
  float osmoothedVal = smoothedPs[i-1]; //???;
  float smpathprob = (1-lam)*osmoothedVal;
  float nsmoothedVal = uspathprob+smpathprob;
  smoothedPs[i] = nsmoothedVal;
  float pm = uspathprob/nsmoothedVal;
  //cerr << "\t" << uspathprob << " " << smpathprob << " " << pm << endl;
  trData[i][b].c++;
  trData[i][b].pm += pm;  
}

void
callProcG(TreeHist* treeh)
{
  FeatureTree* ginfo[MAXNUMFS];
  ginfo[0] = FeatureTree::root(); 
  int cVal = (*Feature::conditionedEvent)(treeh);
  if(cVal < 0) return;
  curTree = treeh->tree; //???;
  //cerr << *curTree << endl;
  for(int i = 1 ; i <= Feature::total[Feature::whichInt] ; i++)
    processG(i, ginfo, treeh, cVal);
}

void
gatherFfCounts(InputTree* tree, int inHpos)
{
  InputTrees& st = tree->subTrees();;
  InputTrees::iterator  subTreeIter= st.begin();
  InputTree  *subTree;
  int hpos = 0;
  if(st.size() != 0) hpos = headPosFromTree(tree);
  //cerr << hpos << *tree << endl;
  int pos = 0;
  for( ; subTreeIter != st.end() ; subTreeIter++ )
    {
      subTree = *subTreeIter;
      gatherFfCounts(subTree, pos==hpos ? 1 : 0);
      pos++;
    }
  //cerr << "g " << *tree << endl;
  curTree = tree;
  TreeHist treeh(tree, 0);
  treeh.pos = pos;
  treeh.hpos = hpos;
  const Term* lhsTerm = Term::get(tree->term());
  int wI = Feature::whichInt;
  if(wI == HCALC || wI == UCALC)
    {
      if(!inHpos) callProcG(&treeh);
      return;
    }
  if(lhsTerm->terminal_p())
    {
      if(wI == TTCALC) callProcG(&treeh);
      return;
    }
  if(st.size() == 1 && st.front()->term() == tree->term()) return;
  //cerr << "gff " << *tree << endl;
  subTreeIter = st.begin();
  int cVal;
  treeh.pos = -1;
  if(wI == LMCALC) callProcG(&treeh);
  if(wI == LCALC) callProcG(&treeh);
  pos = 0;
  for( ; subTreeIter != st.end() ; subTreeIter++)
    {
      treeh.pos = pos;
      if(pos == hpos && wI == MCALC) callProcG(&treeh);
      if(pos < hpos && wI == LCALC) callProcG(&treeh);
      if(pos > hpos && wI == RCALC) callProcG(&treeh);
      if(pos == hpos && wI == RUCALC) callProcG(&treeh);
      if(pos >= hpos && wI == RMCALC) callProcG(&treeh);
      if(pos <= hpos && wI == LMCALC) callProcG(&treeh);
      pos++;
    }
  //cerr << "gg " << *tree << endl;
  treeh.pos = pos;
  if(wI == RCALC) callProcG(&treeh);
  if(wI == RMCALC) callProcG(&treeh);
}

float curLogBases[20];
float totForFeat[20];
float prevMeanSq[20];
float curMeanSq[20];
int done[20] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

int
resetLogBases(int n)
{
  int i,j;
  int stillWorking = 0;
  for(i = 2 ; i <= Feature::total[whichInt] ; i++)
    {
      if(done[i]) continue;
      stillWorking = 1;
      curLogBases[i] += .1;
      curMeanSq[i] = 0;
      totForFeat[i] = 0;
      for(j = 1 ; j < 15 ; j++) totForFeat[i] += trData[i][j].c;
      for(j = 1 ; j < 15 ; j++)
	{
	  curMeanSq[i] += pow((trData[i][j].c - (totForFeat[i]/13.0)),2);
	}
      if((prevMeanSq[i] > 0 && curMeanSq[i] >= prevMeanSq[i])
	 || trData[i][14].c == 0)
	{
	  //cerr << "rlb " << i << " "
	  //   << curLogBases[i]-.1 << " " << curMeanSq[i]
	  //   << " " << trData[i][14].c << " " << prevMeanSq[i] << endl;
	  done[i] = 1;
	  curLogBases[i] -= 0.1;
	}
      prevMeanSq[i] = curMeanSq[i];
    }
  return stillWorking;
}

void
printCounts()
{
   int b,f;
   cout << "\n";
   for(f = 2 ; f <= Feature::total[whichInt] ; f++)
     cout << "\t" << curMeanSq[f];
   cout << "\n";
   for(b = 1; b < 15 ; b++)
     {
       cout << b ;
       for(f = 2; f <= Feature::total[whichInt] ; f++)
	 {
	   cout << "\t" << trData[f][b].c;
	 }
       cout << "\n";
     }
}

void
zeroData()
{
   int b,f;
   for(b = 0; b < 15 ; b++)
     {
       for(f = 1; f <= Feature::total[whichInt] ; f++)
	 {
	   trData[f][b].c = 0;
	   trData[f][b].pm = 0.0;
	 }
     }
}

void
pickLogBases(InputTree** trainingData, int sc)
{
  int i;
  for( i = 0 ; i < 20 ; i++)
    {
      curLogBases[i] = (whichInt == HCALC) ? 1.1 : 1.5 ;
    }
  for(i = 0 ; i < Feature::total[whichInt] ; i++) prevMeanSq[i] = 0;
  for(i = 0 ; ; i++)
    {
      goThroughSents(trainingData,sc);
      int contp = resetLogBases(i);
      if(!contp) break;
      //pprintCounts();
      zeroData();
    }
}  

void
printLambdas(ostream& res)
{
   int b,f;
   for(f = 2 ; f <= Feature::total[whichInt] ; f++)
     res << "\t" << curLogBases[f];
   res << "\n";
   for(b = 1; b < 15 ; b++)
     {
       res << b ;
       for(f = 2; f <= Feature::total[whichInt] ; f++)
	 {
	   res << "\t";
	   float val = Feature::getLambda(whichInt, f, b);
	   if(val > .998) val = .998;  //nothing too close to 1; 
	   res << val;
	 }
       res << "\n";
     }
}

void
updateLambdas()
{
   int b,f;
   for(b = 1; b < 15 ; b++)
     {
       for(f = 2; f <= Feature::total[whichInt] ; f++)
	 {
	   int denom = trData[f][b].c;
	   float num = trData[f][b].pm;
	   float ans = num/(float)denom;
	   if(denom == 0) ans = ((float)b/15.0);
	   else if(denom < 5 && trData[f][b+1].pm == 0 &&
		   (b == 0 ||  trData[f][b-1].pm == 0))
	     ans = 0;
	   //cerr << ans << " " << num << " " << denom << endl;
	   Feature::setLambda(whichInt,f,b,ans);
	 }
     }

   for(b = 2 ; b < 15 ; b++)
     for(f = 2; f <= Feature::total[whichInt] ; f++)
       {
	 int denom = trData[f][b].c;
	 if(denom != 0) continue;
	 float ans;
	 float prevLam = Feature::getLambda(whichInt, f, b-1);
	 float nxtLam = Feature::getLambda(whichInt, f, f+2);
	 ans = (prevLam+nxtLam)/2;
	 Feature::setLambda(whichInt,f,b,ans);
       }

}

void
lamInit()
{
   int b,f;
   float factor = 1.0/15.0;
   int numFs = Feature::total[whichInt];
   for(b = 0; b < 15 ; b++)
     {
       for(f = 1; f <= numFs ; f++)
	 {
	   Feature* feat = Feature::fromInt(f, whichInt); 
	   int strtpos = feat->startPos;
	   /* we are trying b/15 */
	   float num = factor* b;
	   Feature::setLambda(whichInt,f,b,num);
	   trData[f][b].c = 0;
	   trData[f][b].pm = 0.0;
	 }
     }
   //printLambdas(cout);
}

int
main(int argc, char *argv[])
{
   struct rlimit 	core_limits;
   core_limits.rlim_cur = 0;
   core_limits.rlim_max = 0;
   setrlimit( RLIMIT_CORE, &core_limits );

   ECArgs args( argc, argv );
   assert(args.nargs() == 2);
   conditionedType = args.arg(0);
   cerr << "start trainRs: " << conditionedType << endl;

   string  path( args.arg( 1 ) );
   Term::init(path);
   readHeadInfo(path);

   Pst pst(path);

   addSubFeatureFns();
   Feature::init(path, conditionedType); 

   whichInt = Feature::whichInt;
   int ceFunInt = Feature::conditionedFeatureInt[Feature::whichInt];
   Feature::conditionedEvent
     = SubFeature::Funs[ceFunInt];

   Feat::Usage = PARSE;
   string ftstr(path);
   ftstr += conditionedType;
   ftstr += ".g";
   ifstream fts(ftstr.c_str());
   assert(fts);
   tRoot = new FeatureTree(fts); //puts it in root;

   cout.precision(3);
   cerr.precision(3);

   lamInit();

   int sentenceCount = 0;
   int i;
   InputTree* trainingData[1301];
   for(i=0 ; ; i++)
     {
       if(i > 1300) break;
       InputTree*     correct = new InputTree;
       cin >> (*correct);
       if(correct->length() == 0) break;

       EcSPairs wtList;
       correct->make(wtList);
       //cerr << sentenceCount << *par << endl;
       trainingData[i] = correct;
      }

   sentenceCount = i;
   pickLogBases(trainingData,sentenceCount);

   for(pass = 0 ; pass < 10 ; pass++)
     {
       if(pass%2 == 0) cout << "Pass " << pass << endl;
       goThroughSents(trainingData, sentenceCount);
       updateLambdas();
       //printLambdas(cout);
       zeroData();
     }
   string resS(path);
   resS += conditionedType;
   resS += ".lambdas";
   ofstream res(resS.c_str());
   res.precision(3);
   printLambdas(res);
   cout << "Total params = " << FeatureTree::totParams << endl;
   cout << "Done: " << (int)sbrk(0) << endl;
}


void
goThroughSents(InputTree* trainingData[1301], int sc)
{
  int sentenceCount;
  for(sentenceCount = 0 ; sentenceCount < sc ; sentenceCount++)
    {
      InputTree* par = trainingData[sentenceCount];
      //if(sentenceCount%50 == 1)
      //cerr << sentenceCount << endl;
      makeSent(par);
      //cerr << "gts " << endPos << " " << *par << endl;
      gatherFfCounts(par,0);
      if(whichInt == TTCALC)
	{
	  list<InputTree*> dummy2;
	  InputTree stopInputTree(par->finish(),par->finish(),
				  whichInt==TTCALC ? "" : "^^",
				  "STOP","",
				  dummy2,NULL,NULL);
	  stopInputTree.headTree() = &stopInputTree;
	  TreeHist treeh(&stopInputTree,0);
	  treeh.hpos = 0;
	  callProcG(&treeh);
	}
    }
}
