#include "TreeHist.h"
#include "treeHistSf.h"
#include "Term.h"
#include "Feature.h"
#include "InputTree.h"
#include "Pst.h"
#include "headFinder.h"
#include "Feat.h"
#include "ccInd.h"

int
tree_term(TreeHist* treeh)
{
  InputTree* tree = treeh->tree;
  const ECString& trmStr  = tree->term();
  const Term* trm = Term::get(trmStr);
  assert(trm);
  return trm->toInt();
}

int
tree_parent_term(TreeHist* treeh)
{
  InputTree* tree = treeh->tree;
  static int s1int = 0;
  if(!s1int)
    {
      ECString s1nm("S1");
      s1int = Term::get(s1nm)->toInt();
    }
  InputTree* par = tree->parent();
  if(!par) return s1int;
  const ECString& trmStr  = par->term();
  const Term* trm = Term::get(trmStr);
  assert(trm);
  assert(!trm->terminal_p());
  return trm->toInt();
}

int
toBe(const ECString& parw)
{
  if(parw == "was" || parw == "is" || parw == "be" || parw == "been"
     || parw == "are" || parw == "were" || parw == "being")
    return 1;
  else return 0;
}

int
tree_parent_pos(TreeHist* treeh)
{
  static int stopint = 0;
  InputTree* tree = treeh->tree;
  if(!stopint)
    {
      ECString stopnm("STOP");
      stopint = Term::get(stopnm)->toInt();
    }
  InputTree* par = tree->parent();
  if(!par) return stopint;

  const ECString& trmStr  = par->hTag();
  const Term* trm = Term::get(trmStr);
  assert(trm);
  if(!trm->terminal_p())
    {
      cerr << "Bad head Part of Speech: " << *trm << " in " <<endl;
      cerr << *tree << endl;
      assert(trm->terminal_p());
    }
  int ans = trm->toInt();
  if(ans < 2 && toBe(par->head())) return 48;
  return ans;
}

int
tree_term_before(TreeHist* treeh)
{
  static int stopint = 0;
  if(!stopint)
    {
      ECString stopnm("STOP");
      stopint = Term::get(stopnm)->toInt();
    }
  InputTree* tree = treeh->tree;
  InputTree* par = tree->parent();
  if(!par) return stopint;
  InputTreesIter iti = par->subTrees().begin();
  for( ; iti != par->subTrees().end() ; iti++ )
    {
      InputTree* st = *iti;
      if(st != tree) continue;
      if(iti == par->subTrees().begin()) return stopint;
      iti--;
      st = *iti;
      const ECString& trmStr  = st->term();
      const Term* trm = Term::get(trmStr);
      assert(trm);
      return trm->toInt();
    }
  error("Should never get here");
  return -1;
}

int
tree_term_after(TreeHist* treeh)
{
  static int stopint = 0;
  if(!stopint)
    {
      ECString stopnm("STOP");
      stopint = Term::get(stopnm)->toInt();
    }
  InputTree* tree = treeh->tree;
  InputTree* par = tree->parent();
  if(!par) return stopint;
  InputTreesIter iti = par->subTrees().begin();
  for( ; iti != par->subTrees().end() ; iti++ )
    {
      InputTree* st = *iti;
      if(st != tree) continue;
      iti++;
      if(iti == par->subTrees().end()) return stopint;
      st = *iti;
      const ECString& trmStr  = st->term();
      const Term* trm = Term::get(trmStr);
      assert(trm);
      return trm->toInt();
    }
  error("Should never get here");
  return -1;
}

int
tree_pos(TreeHist* treeh)
{
  InputTree* tree = treeh->tree;
  const ECString& trmStr  = tree->hTag();
  const Term* trm = Term::get(trmStr);
  assert(trm);
  if(!trm->terminal_p())
    {
      cerr << "Bad head Part of Speech: " << *trm << " in " <<endl;
      cerr << *tree << endl;
      assert(trm->terminal_p());
    }
  return trm->toInt();
}


int
tree_head(TreeHist* treeh)
{
  InputTree* tree = treeh->tree;
  char temp[128];
  ECString wrdStr(toLower(tree->head().c_str(), temp));
  const WordInfo* wi = Pst::get(wrdStr);
  if(!wi)
    {
      if(Feat::Usage == PARSE) return -1;
      assert(wi);
    }
  int ans = wi->toInt();
  assert(ans >= 0);
  return ans;
}

int
tree_parent_head(TreeHist* treeh)
{
  InputTree* tree = treeh->tree;
  InputTree* pt = tree->parent();
  
  if(!pt)
    {
      ECString temp("^^");
      return Pst::get(temp)->toInt();
    }
  char temp[128];
  ECString wrdStr(toLower(pt->head().c_str(),temp));
  const WordInfo* wi = Pst::get(wrdStr);
  if(!wi)
    {
      if(Feat::Usage == PARSE) return -1;
      assert(wi);
    }
  int ans = wi->toInt();
  assert(ans >= 0);
  return ans;
}

int
tree_true(TreeHist* treeh)
{
  return 1;
}

int
tree_grandparent_term(TreeHist* treeh)
{
  static int s1int = 0;
  if(!s1int)
    {
      ECString s1nm("S1");
      s1int = Term::get(s1nm)->toInt();
    }
  InputTree* tree = treeh->tree;
  InputTree* par = tree->parent();
  if(!par) return s1int;
  InputTree* gpar = par->parent();
  if(!gpar) return s1int;
  const ECString& trmStr  = gpar->term();
  const Term* trm = Term::get(trmStr);
  assert(trm);
  assert(!trm->terminal_p());
  return trm->toInt();
}

int
tree_grandparent_pos(TreeHist* treeh)
{
  static int stopint = 0;
  if(!stopint)
    {
      ECString stopnm("STOP");
      stopint = Term::get(stopnm)->toInt();
    }
  InputTree* tree = treeh->tree;
  InputTree* par1 = tree->parent();
  if(!par1) return stopint;
  InputTree* par = par1->parent();
  if(!par) return stopint;
  
  const ECString& trmStr  = par->hTag();
  const Term* trm = Term::get(trmStr);
  assert(trm);
  if(!trm->terminal_p())
    {
      cerr << "Bad head Part of Speech: " << *trm << " in " <<endl;
      cerr << *tree << endl;
      assert(trm->terminal_p());
    }
  return trm->toInt();
}

int
tree_grandparent_head(TreeHist* treeh)
{
  InputTree* tree = treeh->tree;
  InputTree* pt = tree->parent();
  static int topInt = -1;
  if(topInt < 0)
    {
      ECString temp("^^");
      topInt = Pst::get(temp)->toInt();
    }
  if(!pt) return topInt;
  pt = pt->parent();
  if(!pt) return topInt;

  char temp[128];
  ECString wrdStr(toLower(pt->head().c_str(),temp));
  const WordInfo* wi = Pst::get(wrdStr);
  if(!wi)
    {
      if(Feat::Usage == PARSE) return -1;
      assert(wi);
    }
  int ans = wi->toInt();
  assert(ans >= 0);
  return ans;
}

int
tree_ccparent_term(TreeHist* treeh)
{
  static int s1int = 0;
  if(!s1int)
    {
      ECString s1nm("S1");
      s1int = Term::get(s1nm)->toInt();
    }
  assert(treeh);
  InputTree* tree = treeh->tree;
  assert(tree);
  InputTree* par = tree->parent();
  if(!par) return s1int;
  const ECString& trmStr  = par->term();
  const Term* trm = Term::get(trmStr);
  assert(trm);
  int trmInt = trm->toInt();
  if(trmStr != tree->term()) return trmInt; //??? new;
  assert(!trm->terminal_p());
  //int ccedtrmInt = ltocc(trmInt,ccIndFromTree(par));
  int ccedtrmInt = ccIndFromTree(par);
  return ccedtrmInt;
}

int
tree_cc_ind(TreeHist* treeh)
{
  InputTree* tree = treeh->tree;
  InputTree* par = tree->parent();
  if(!par) return 0; // 0 ==> no CC;
  const ECString& trmStr  = par->term();
  const Term* trm = Term::get(trmStr);
  assert(trm);
  int trmInt = trm->toInt();
  if(trmStr != tree->term()) return 0;
  assert(!trm->terminal_p());
  int ccedInt = ccIndFromTree(par);
  return ccedInt;
}

int
tree_size(TreeHist* treeh)
{
  static int bucs[9] = {1, 3, 6, 10, 15, 21, 28, 36, 999};
  InputTree* tree = treeh->tree;
  int sz = tree->finish() - tree->start();
  for(int i = 0 ; i < 9 ; i++)
    if(sz <= bucs[i]) return i;
  assert("Never get here");
  return -1;
}


extern InputTree* sentence[256];
extern int endPos;

int
tree_effEnd(TreeHist* treeh)
{
  InputTree* tree = treeh->tree;
  int pos = tree->finish();
  bool ans;
  if(pos > endPos)
    {
      cout << "Pos > endPos" << endl;
      ans = 0;
    }
  else if(pos == endPos) ans = 1;
  else
    {
      ECString wrd = sentence[pos]->word();
      ECString trm = sentence[pos]->term();
      if(trm == "." || wrd == ";") ans = 1;
      else if((pos+2) > endPos) ans = 0;
      else if(wrd == ",")
	{
	  if(sentence[pos+1]->word() == "''")
	    ans = 1; // ,'' acts like end of sentence;
	  else ans = 0;  //ans = 2 for alt version???
			     }
      else ans = 0;
    }
  return ans;
}

int
tree_ngram(TreeHist* treeh, int n, int l)
{
  static int stopTermInt = -1;
  if(stopTermInt < 0)
    {
      ECString stopStr("STOP");
      const Term* stopTerm = Term::get(stopStr);
      stopTermInt = stopTerm->toInt();
    }

  int pos = treeh->pos;
  int hp = treeh->hpos;
  int m = pos + (n * l);
  if(m < 0) return stopTermInt;
  InputTree* tree = treeh->tree;
  if(m >= tree->subTrees().size()) return stopTermInt;
  if(m > hp && l > 0) return stopTermInt;
  InputTree  *subTree;
  InputTreesIter  subTreeIter = tree->subTrees().begin();
  int i = 0;
  for( ; subTreeIter != tree->subTrees().end() ; subTreeIter++ )
    {
      if(i == m)
	{
	  subTree = *subTreeIter;
	  const Term* trm = Term::get(subTree->term());
	  return trm->toInt();
	}
      i++;
    }
  assert("should never get here");
  return -1;
}

InputTree*
tree_find(TreeHist* treeh, int n)
{
  int pos = treeh->pos;
  int hp = treeh->hpos;
  int m = pos + n;
  assert(m >= 0);
  InputTree* tree = treeh->tree;
  assert(!(m >= tree->subTrees().size()));
  InputTree  *subTree;
  InputTreesIter  subTreeIter = tree->subTrees().begin();
  int i = 0;
  for( ; subTreeIter != tree->subTrees().end() ; subTreeIter++ )
    {
      if(i == m)
	{
	  subTree = *subTreeIter;
	  return subTree;
	}
      i++;
    }
  assert("should never get here");
  return NULL;
}

int
tree_left0(TreeHist* treeh)
{
  return tree_ngram(treeh, 0, 0);
}

int
tree_left1(TreeHist* treeh)
{
  return tree_ngram(treeh, 1, 1);
}

int
tree_left2(TreeHist* treeh)
{
  return tree_ngram(treeh, 2, 1);
}

int
tree_left3(TreeHist* treeh)
{
  return tree_ngram(treeh, 3, 1);
}

int
tree_right1(TreeHist* treeh)
{
  return tree_ngram(treeh, 1, -1);
}

int
tree_right2(TreeHist* treeh)
{
  return tree_ngram(treeh, 2, -1);
}

int
tree_right3(TreeHist* treeh)
{
  return tree_ngram(treeh, 3, -1);
}

int
tree_right4(TreeHist* treeh)
{
  return tree_ngram(treeh, 4, -1);
}

int
tree_left4(TreeHist* treeh)
{
  return tree_ngram(treeh, 4, 1);
}

int
tree_noopenQr(TreeHist* treeh)
{
  InputTree* tree = treeh->tree;
  int pos = treeh->pos;
  int sz = tree->subTrees().size();
  InputTree  *subTree;
  InputTreesIter  subTreeIter = tree->subTrees().begin();
  int i = 0;
  bool sawOpen = false;
  
  for( ; ; subTreeIter++ )
    {
      if(i == pos) break; 
      subTree = *subTreeIter;
      assert(i < sz);
      const Term* trm = Term::get(subTree->term());
      if(subTree->term() == "``") sawOpen = true;
      if(subTree->term() == "''") sawOpen = false;
      i++;
    }
  if(sawOpen) return 0;
  else return 1;
}

int
tree_noopenQl(TreeHist* treeh)
{
  InputTree* tree = treeh->tree;
  int pos = treeh->pos;
  int hpos = treeh->hpos;
  InputTree  *subTree;
  InputTrees::reverse_iterator  subTreeIter = tree->subTrees().rbegin();
  int i = tree->subTrees().size()-1;
  bool sawOpen = false;
  bool sawClosed = false;
  
  for( ; ; subTreeIter++ )
    {
      if(i == pos) break;
      if(i > hpos) { i-- ; continue; }
      assert(i >= 0);
      subTree = *subTreeIter;
      const Term* trm = Term::get(subTree->term());
      if(subTree->term() == "''") sawOpen = true;
      else if(subTree->term() == "``") sawOpen = false;
      i--;
    }
  if(sawOpen) return 0;
  else return 1;
}

int
tree_B(TreeHist* treeh, int blInd)
{
  InputTree* tree = treeh->tree;
  int i;
  int pos = treeh->pos;
  int hpos = treeh->hpos;
  //cerr << "tb1 " << pos << " " << hpos << " " << *tree << endl;
  int sz = tree->subTrees().size();
  int wpos;
  assert(pos <= sz);
  //cerr << "tb " << pos << " " << hpos << " " << sz << endl;
  if(pos < 0) wpos = tree->start()-1;
  else if(sz == 0) wpos = tree->start()-1;
  else if(pos == sz) wpos = tree->finish();
  else
    {
      InputTreesIter iti = tree->subTrees().begin();
      i = 0;
      for( ; iti != tree->subTrees().end() ; iti++)
	{
	  if(i < pos) {i++; continue;}
	  InputTree* st = *iti;
	  if(pos < hpos) wpos = st->start()-1;
	  else if(pos > hpos) wpos = st->finish();
	  else if(blInd) wpos = st->start()-1;
	  else wpos = st->finish();
	  //cerr << "tbf " << *st << " " << wpos << endl;
	  break;
	}
    }
  //cerr << "tb2 " << wpos << endl;
  assert(wpos <= endPos);
  if(wpos < 0 || wpos == endPos) return Term::stopTerm->toInt();
  else return Term::get(sentence[wpos]->term())->toInt();
}

int
tree_Bl(TreeHist* treeh)
{
  return tree_B(treeh, 1);
}

int
tree_Br(TreeHist* treeh)
{
  return tree_B(treeh, 0);
}


int
tree_knownhead(TreeHist* treeh)
{
  static int boundaries[5] = {1, 4, 16, 64, 256};
  InputTree* tree = treeh->tree;
  char temp[128];
  ECString wrdStr(toLower(tree->head().c_str(), temp));
  const WordInfo* wi = Pst::get(wrdStr);
  if(!wi)
    {
      if(Feat::Usage == PARSE) return -1;
      assert(wi);
    }
  int cnt = wi->c();
  int ans;
  if(cnt > 1) return 1;
  else return 0;
  /*
  for(ans = 0 ; ans < 5 ; ans++)
    {
      if(cnt <= boundaries[ans]) break;
    }
  assert(ans >= 0);
  return ans;
  */
}

void
addSubFeatureFns()
{
  /*
    0 t  tree_term
    1 l  tree_parent_term
    2 u  tree_pos
    3 h  tree_head
    4 i  tree_parent_head
    5 T  tree_true
    6 v  tree_parent_pos
    7 b  tree_term_before
    8 a  tree_term_after
    9 m  tree_grandparent_term
    10 w tree_grandparent_pos
    11 j tree_grandparent_head
    12 c tree_ccparent_term
    13 L1 tree_left1
    14 L2 tree_left2
    15 R1 tree_right1
    16 R2 tree_right2
    17 Qr tree_noopenQr
    18 L0 tree_left0;
    19 L3 tree_left3
    20 R3 tree_right3
    21 Ql tree_noopenQl
    22 Bl tree_Bl
    23 Br tree_Br
    24 e  tree_effEnd
    25 tr tree_true
    */
  int (*funs[26])(TreeHist*)
    = {tree_term, tree_parent_term, tree_pos, tree_head, tree_parent_head,
       tree_true, tree_parent_pos, tree_term_before,
       tree_term_after, tree_grandparent_term, tree_grandparent_pos,
       tree_grandparent_head, tree_ccparent_term,
       tree_left1, tree_left2, tree_right1, tree_right2, tree_noopenQr,
       tree_left0, tree_left3, tree_right3,tree_noopenQl,tree_Bl,tree_Br,
       tree_effEnd, tree_true
    };
  int i;
  for(i = 0 ; i < 26 ; i++)
      SubFeature::Funs[i] = funs[i];
}
